@@ -41,10 +41,13 @@ for testing, patches are welcome.
 
 See the README file for an example usage of this module.
 
+This module supports Python 2.6 and later.
+
 """
 
 
-__version__ = 'lowell87@gmail.com-20130711013447-67x7gyy94jtobw0v 2013-07-10 21:34:47 -0400 [1]'
+__version__  = '0.9.0'
+__revision__  = 'lowell87@gmail.com-20130706224727-2s0okpzxwye0fxop 2013-07-06 18:47:27 -0400 [0]'
 __author__ = "Lowell Alleman"
 __all__ = [
     "ConcurrentRotatingFileHandler",
@@ -54,7 +57,7 @@ __all__ = [
 import os
 import sys
 from random import randint
-from logging import Handler, LogRecord
+from logging import Handler
 from logging.handlers import BaseRotatingHandler
 
 try:
@@ -72,18 +75,6 @@ except ImportError:
 # sibling module than handles all the ugly platform-specific details of file locking
 from portalocker import lock, unlock, LOCK_EX, LOCK_NB, LockException
 
-
-# A client can set this to true to automatically convert relative paths to
-# absolute paths (which will also hide the absolute path warnings)
-FORCE_ABSOLUTE_PATH = False
-
-# Workaround for handleError() in Python 2.7+ where record is written to stderr
-class NullLogRecord(LogRecord):
-    def __init__(self):
-        pass
-    def __getattr__(self, attr):
-        return None
-
 class ConcurrentRotatingFileHandler(BaseRotatingHandler):
     """
     Handler for logging to a set of files, which switches from one file to the
@@ -92,7 +83,7 @@ class ConcurrentRotatingFileHandler(BaseRotatingHandler):
     exceed the given size.
     """
     def __init__(self, filename, mode='a', maxBytes=0, backupCount=0,
-                 encoding=None, debug=True, supress_abs_warn=False):
+                 encoding=None, debug=True, delay=0):
         """
         Open the specified file and use it as the stream for logging.
 
@@ -121,8 +112,9 @@ class ConcurrentRotatingFileHandler(BaseRotatingHandler):
         the other processes need to close the file first.  A mechanism, called
         "degraded" mode, has been created for this scenario.  In degraded mode,
         the log file is closed after each log message is written.  So once all
-        processes have entered degraded mode, the next rotate log attempt should
-        be successful and then normal logging can be resumed.
+        processes have entered degraded mode, the net rotation attempt should
+        be successful and then normal logging can be resumed.  Using the 'delay'
+        parameter may help reduce contention in some usage patterns.
 
         This log handler assumes that all concurrent processes logging to a
         single file will are using only this class, and that the exact same
@@ -131,110 +123,85 @@ class ConcurrentRotatingFileHandler(BaseRotatingHandler):
         different values for 'maxBytes' or 'backupCount', then odd behavior is
         expected. The same is true if this class is used by one application, but
         the RotatingFileHandler is used by another.
-
-        NOTE:  You should always provide 'filename' as an absolute path, since
-        this class will need to re-open the file during rotation. If your
-        application call os.chdir() then subsequent log files could be created
-        in the wrong directory.
         """
-        # The question of absolute paths: I'm not sure what the 'right thing' is
-        # to do here. RotatingFileHander simply ignores this possibility. I was
-        # going call os.path.abspath(), but that potentially limits uses. For
-        # example, on Linux (any posix system?) you can rename a directory of a
-        # running app, and the app wouldn't notice as long as it only opens new
-        # files using relative paths. But since that's not a "normal" thing to
-        # do, and having an app call os.chdir() is a much more likely scenario
-        # that should be supported. For the moment, we are just going to warn
-        # the user if they provide a relative path and do some other voodoo
-        # logic that you'll just have to review for yourself.
-        
-        # if the given filename contains no path, we make an absolute path
-        if not os.path.isabs(filename):
-            if FORCE_ABSOLUTE_PATH or \
-               not os.path.split(filename)[0]:
-                filename = os.path.abspath(filename)
-            elif not supress_abs_warn:
-                from warnings import warn
-                warn("The given 'filename' should be an absolute path.  If your "
-                     "application calls os.chdir(), your logs may get messed up. "
-                     "Use 'supress_abs_warn=True' to hide this message.")
-        try:
-            BaseRotatingHandler.__init__(self, filename, mode, encoding)
-        except TypeError: # Due to a different logging release without encoding support  (Python 2.4.1 and earlier?)
-            BaseRotatingHandler.__init__(self, filename, mode)
-            self.encoding = encoding
-        
+        # Absolute file name handling done by FileHandler since Python 2.5  
+        BaseRotatingHandler.__init__(self, filename, mode, encoding, delay)
+        self.delay = delay
         self._rotateFailed = False
         self.maxBytes = maxBytes
         self.backupCount = backupCount
-        # Use 'file.lock' and not 'file.log.lock' (Only handles the normal "*.log" case.)
-        if filename.endswith(".log"):
-            lock_file = filename[:-4]
-        else:
-            lock_file = filename
-        self.stream_lock = open(lock_file + ".lock", "w")
-        
+        self._open_lockfile()
         # For debug mode, swap out the "_degrade()" method with a more a verbose one.
         if debug:
             self._degrade = self._degrade_debug
     
-    def _openFile(self, mode):
-        if self.encoding:
-            self.stream = codecs.open(self.baseFilename, mode, self.encoding)
+    def _open_lockfile(self):
+        # Use 'file.lock' and not 'file.log.lock' (Only handles the normal "*.log" case.)
+        if self.baseFilename.endswith(".log"):
+            lock_file = self.baseFilename[:-4]
         else:
-            self.stream = open(self.baseFilename, mode)
+            lock_file = self.baseFilename
+        lock_file += ".lock"
+        self.stream_lock = open(lock_file,"w")
+    
+    def _open(self, mode=None):
+        """
+        Open the current base file with the (original) mode and encoding.
+        Return the resulting stream.
+        
+        Note:  Copied from stdlib.  Added option to override 'mode'
+        """
+        if mode is None:
+            mode = self.mode
+        if self.encoding is None:
+            stream = open(self.baseFilename, mode)
+        else:
+            stream = codecs.open(self.baseFilename, mode, self.encoding)
+        return stream
+    
+    def _close(self):
+        """ Close file stream.  Unlike close(), we don't tear anything down, we
+        expect the log to be re-opened after rotation."""
+        if self.stream:
+            try:
+                if not self.stream.closed:
+                    # Flushing probably isn't technically necessary, but it feels right
+                    self.stream.flush()
+                    self.stream.close()
+            finally:
+                self.stream = None
     
     def acquire(self):
         """ Acquire thread and file locks. Also re-opening log file when running
         in 'degraded' mode. """
         # handle thread lock
         Handler.acquire(self)
+        # Issue a file lock.  (This is inefficient for multiple active threads
+        # within a single process. But if you're worried about high-performance,
+        # you probably aren't using this log handler.)
         lock(self.stream_lock, LOCK_EX)
-        if self.stream.closed:
-            self._openFile(self.mode)
-    
+        # Stream will be opened as part by FileHandler.emit()
+
     def release(self):
-        """ Release file and thread locks. Flush stream and take care of closing
-        stream in 'degraded' mode. """
+        """ Release file and thread locks. If in 'degraded' mode, close the
+        stream to reduce contention until the log files can be rotated. """
         try:
-            try:
-                if not self.stream.closed:
-                    self.stream.flush()
-                    if self._rotateFailed:
-                        self.stream.close()
-            except Exception:
-                self.handleError(NullLogRecord())
+            if self._rotateFailed:
+                self._close()
         finally:
             try:
-                if not self.stream_lock.closed:
-                    unlock(self.stream_lock)
+                unlock(self.stream_lock)
             finally:
                 # release thread lock
                 Handler.release(self)
     
     def close(self):
         """
-        Closes the stream.
-        """
-        if not self.stream.closed:
-            self.stream.flush()
-            self.stream.close()
-        if not self.stream_lock.closed:
-            self.stream_lock.close()
-        Handler.close(self)
-    
-    def flush(self):
-        """ flush():  Do nothing.
-
-        Since a flush is issued in release(), we don't do it here. To do a flush
-        here, it would be necessary to re-lock everything, and it is just easier
-        and cleaner to do it all in release(), rather than requiring two lock
-        ops per handle() call.
-
-        Doing a flush() here would also introduces a window of opportunity for
-        another process to write to the log file in between calling
-        stream.write() and stream.flush(), which seems like a bad thing. """
-        pass
+        Closes the stream.  Note that we don't close stream_lock. """
+        try:
+            self._close()    
+        finally:
+            Handler.close(self)
     
     def _degrade(self, degrade, msg, *args):
         """ Set degrade mode or not.  Ignore msg. """
@@ -260,14 +227,16 @@ class ConcurrentRotatingFileHandler(BaseRotatingHandler):
         """
         Do a rollover, as described in __init__().
         """
+        self._close()
         if self.backupCount <= 0:
             # Don't keep any backups, just overwrite the existing backup file
             # Locking doesn't much matter here; since we are overwriting it anyway
-            self.stream.close()
-            self._openFile("w")
+            self.stream = self._open("w")
             return
-        self.stream.close()
         try:
+            # Determine if we can rename the log file or not. Windows refuses to
+            # rename an open file, Unix is inode base so it doesn't care.
+            
             # Attempt to rename logfile to tempname:  There is a slight race-condition here, but it seems unavoidable
             tmpname = None
             while not tmpname or os.path.exists(tmpname):
@@ -282,12 +251,14 @@ class ConcurrentRotatingFileHandler(BaseRotatingHandler):
                 return
             
             # Q: Is there some way to protect this code from a KeboardInterupt?
-            # This isn't necessarily a data loss issue, but it certainly would
-            # break the rotation process during my stress testing.
+            # This isn't necessarily a data loss issue, but it certainly does 
+            # break the rotation process during stress testing.
             
             # There is currently no mechanism in place to handle the situation
             # where one of these log files cannot be renamed. (Example, user
-            # opens "logfile.3" in notepad)
+            # opens "logfile.3" in notepad); we could test rename each file, but
+            # nobody's complained about this being an issue; so the additional
+            # code complexity isn't warranted.
             for i in range(self.backupCount - 1, 0, -1):
                 sfn = "%s.%d" % (self.baseFilename, i)
                 dfn = "%s.%d" % (self.baseFilename, i + 1)
@@ -303,7 +274,11 @@ class ConcurrentRotatingFileHandler(BaseRotatingHandler):
             #print "%s -> %s" % (self.baseFilename, dfn)
             self._degrade(False, "Rotation completed")
         finally:
-            self._openFile(self.mode)
+            # Re-open the output stream, but if "delay" is enabled then wait
+            # until the next emit() call. This could reduce rename contention in
+            # some usage patterns.
+            if not self.delay:
+                self.stream = self._open()
     
     def shouldRollover(self, record):
         """
@@ -314,12 +289,17 @@ class ConcurrentRotatingFileHandler(BaseRotatingHandler):
         the file size under maxBytes we ignore the length of the current record.
         """
         del record  # avoid pychecker warnings
+        # Is stream is not yet open, skip rollover check. (Check will occur on
+        # next message, after emit() calls _open())
+        if self.stream is None:
+            return False
         if self._shouldRollover():
-            # if some other process already did the rollover we might
-            # checked log.1, so we reopen the stream and check again on
-            # the right log file
-            self.stream.close()
-            self._openFile(self.mode)
+            # If some other process already did the rollover (which is possible
+            # on Unix) the file our stream may now be named "log.1", thus
+            # triggering another rollover. Avoid this by closing and opening
+            # "log" again.
+            self._close()
+            self.stream = self._open()
             return self._shouldRollover()
         return False
     
